# coding: utf-8

#---------------------------------------------------------------
#
#	オートコンプリートを行う
#
#		cmpl = Complete.new(['X1', 'X1C', 'X1Cs', 'X1Ck', 'X1turbo', 'X68000'])
#
#		p cmpl.select('X1')
#		=> ["X1", "X1C", "X1Cs", "X1Ck", "X1turbo"]
#
#		p cmpl.choice('X1C')
#		=> false
#
#		p cmpl.choice('X1t')
#		=> "X1turbo"
#
class Complete

	def initialize(array)
		@catalog = ([''] + array + ['']).join('!!')
	end

	def select(hint)
		@catalog.scan(/!(#{hint}.*?)!/).flatten
	end

	def choice(hint)
		(it = select(hint)).size == 1 ? it.first : false
	end
end

#---------------------------------------------------------------
#
#	上演する
#
class Kugutsu

	$LOAD_PATH.unshift(File.dirname(__FILE__)).unshift('.')
	load 'kugutsu.conf'; $LOAD_PATH.shift(2)					# load @@configs

	attr_reader :target, :result, :results

	def initialize(target_hint)

		complete = Complete.new(@@configs[:TARGETS].keys)
		unless(target_name = complete.choice(target_hint))
			puts('=== defined targets ==')
			@@configs[:TARGETS].keys.each {|target_name|
				puts('[ %s ]' % target_name)
			}
			raise("target hint didn't match unique one.")
		end
		@target = @@configs[:TARGETS][target_name]

		@ssh_command =  ['ssh']
		@ssh_command += @target[:SSH_OPTIONS] if(@target[:SSH_OPTIONS])
		@ssh_command << '%s@%s' % [@target[:USER], @target[:HOST]]

		@scp_command =  ['scp']
		@scp_command += @target[:SCP_OPTIONS] if(@target[:SCP_OPTIONS])
		@scp_command << '%s@%s' % [@target[:USER], @target[:HOST]]

		set_exec_pause(0)
		set_log_mode(true)
	end

	def set_exec_pause(sec)
		@exec_pause = sec
	end

	@@log_style = {
		:TAB		=> "\t%s",
		:BLACK		=> "\e[30m\e[1;1m%s\e[0m",
		:RED		=> "\e[31m\e[1;1m%s\e[0m",
		:GREEN		=> "\e[32m\e[1;1m%s\e[0m",
		:YELLOW		=> "\e[33m\e[1;1m%s\e[0m",
		:BLUE		=> "\e[34m\e[1;1m%s\e[0m",
		:MAGENTA	=> "\e[35m\e[1;1m%s\e[0m",
		:CYAN		=> "\e[36m\e[1;1m%s\e[0m",
		:WHITE		=> "\e[37m\e[1;1m%s\e[0m",
		:ULINE		=> "\e[4m\e[1;1m%s\e[0m",
	}
	def set_log_mode(withfile)
		@filelog = withfile
	end
	def log(message, type = nil)
		if(@filelog)
			if(!@logfile)
				File.unlink(logfilelink = 'kugutsu-latest.log') rescue true
				(@logfile = open((logfilename = 'kugutsu-%s-%s.log' % [@target[:HOST], Time.now.strftime('%Y%m%d%H%M%S')]), 'w')).sync = true
				File.symlink(logfilename, logfilelink)
			end
			puts((@@log_style[type] || '%s') % message)
			@logfile.puts(message)
		end
	end

	# ssh 公開鍵を登録する
	def regist_ssh_public_key(key_file = '~/.ssh/id_rsa.pub')

		public_key = "'%s'" % open(File.expand_path(key_file)).readline.chomp
		command =  'mkdir -p ~/.ssh; chmod go-rwx ~/.ssh; '
		command << 'grep %s ~/.ssh/authorized_keys || echo %s >> ~/.ssh/authorized_keys; ' % [public_key, public_key]
		command << 'chmod go-rwx ~/.ssh/authorized_keys'
		exec([command])
	end

	def exec0(ssh_command, command, exec_pause)
		@result = []; IO.popen('-') {|io|
			unless(io)
				sleep(exec_pause)
				system(*ssh_command)
				exit($?.exitstatus || false)
			else
				log("\n===> %s" % command.join(' '), :YELLOW)
				io.each {|line|
					@result << line.chomp
					log(line.chomp, :TAB)
				}
			end
		}
		log('---> %s' % (it = $?.exitstatus), it == 0 ? :BLUE : :RED)
		it == 0 or raise 'failed.'
	end

	def exec(command, exec_pause = @exec_pause)
		exec0(@ssh_command + command, command, exec_pause = @exec_pause)
	end

	def scp_export(source, dest, exec_pause = @exec_pause)
		scp_command = @scp_command.dup
		userhost = scp_command.pop
		scp_command << source
		scp_command << '%s:%s' % [userhost, dest]
		exec0(scp_command, %w!cat <<'EOL'! << dest, exec_pause = @exec_pause)
		exec(%w!cat! << dest) rescue true
	end

	def scp_import(source, dest, exec_pause = @exec_pause)
		scp_command = @scp_command.dup
		userhost = scp_command.pop
		scp_command << '%s:%s' % [userhost, source]
		scp_command << dest
		exec0(scp_command, %w!scp! << source << 'your@host:', exec_pause = @exec_pause)
	end

	# コマンドを順に連続実行し、失敗した時点で終了
	def execs(commands)
		@results = []; commands.each {|command|
			exec(command)
			@results << @result
		}
	end

	def exist(filename)
		result = true; exec(%w![ -e! << filename << ']') rescue result = false
		result
	end

	def not_exist(filename)
		result = true; exec(%w|[ ! -e| << filename << ']') rescue result = false
		result
	end

	#	example of opt
	# 
	#	 	in line 10~20	'10,20'
	# 		first			'1,/pattern/'
	# 		in block		'/begin_pat/,/end_pat/'
	# 		last block		'/begin_pat/,$'
	# 
	def ss_comment_out(param, mark = '#', opt = '')
		"'#{opt}s!^#?(#{param})!#{mark}\\1!'"
	end

	def ss_uncomment(param, mark = '#', opt = '')
		"'#{opt}s!^#{mark}?(#{param})!\\1!'"
	end

	def ss_set_param(param, value, opt = '')
		"'#{opt}s!^#?(#{param}).*!\\1#{value}!'"
	end

	def ss_modify(from, to, opt = '')
		"'#{opt}s!#{from}!#{to}!'"
	end

	def ss_exist_param(param)									# append only
		"'^#?(#{param})'"
	end

	def systemctl(params)
		unless(@systemctl)
			@systemctl = %w!systemctl!; exec(%w!which systemctl!) rescue @systemctl = %w!service!
		end		
		@systemctl + (@systemctl[0] == 'systemctl' ? params : params.reverse)
	end

	def dnf(params)
		unless(@dnf)
			@dnf = %w!dnf!; exec(%w!which dnf!) rescue @dnf = %w!yum!
		end		
		@dnf + params
	end

	def orgfilename(filename, style)
		df = filename.split('/')
		df[-1] = style % df[-1]
		df.join('/')
	end

	def backup(filename, style = '%s.org')
		if(not_exist(org = orgfilename(mod = filename, style)))
			execs([
				%w!mv! << mod << org,
				%w!cp -a! << org << mod,
			])
		end
		org
	end

	# 指定したファイルの一部を修正する
	def modify(filename, scripts, style = '%s.org')
		org = backup(filename, style)

		command = %w!sed -r -i!
		command << filename
		while(script = scripts.shift)
			command << '-e' << script
		end
		exec(command)

		exec(%w!diff -bc! << org << filename) rescue true
	end

	# ファイルに記述がなければ追加する
	def append(filename, pattern, lines, style = '%s.org')
		begin
			pattern or raise									# 指定なしなら無条件に追加
			exec(%w!grep -E! << pattern << filename)
		rescue
			scripts = []; lines.each {|line|
				scripts << "'$ a#{line}'"
			}
			modify(filename, scripts, style)
		end
	end

	# ファイルが存在しなければ生成する
	def create(filename, contents, force = false)
		if(not_exist(filename) or force)						# force 指定なら無条件に上書きする
			open(tempfilename = '%s/kugutsu-create-%s' % [@@configs[:TEMPDIR], @target[:HOST]], 'w') {|fh|
				fh.write(contents)
			}
			scp_export(tempfilename, filename)
		end
	end

	# ファイルを持ってくる
	def import(filename)
		importfilename = ('kugutsu-import-%s-%s' % [@target[:HOST], filename]).gsub(/[^-\w.]/, '_')
		scp_import(filename, importfilename)
	end
end

__END__

